# Copyright (c) MONAI Consortium
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import unittest

import numpy as np
import torch

from monai.data import MetaTensor
from monai.transforms import EnsureTyped
from tests.utils import assert_allclose


class TestEnsureTyped(unittest.TestCase):

    def test_array_input(self):
        test_datas = [np.array([[1, 2], [3, 4]]), torch.as_tensor([[1, 2], [3, 4]])]
        if torch.cuda.is_available():
            test_datas.append(test_datas[-1].cuda())
        for test_data in test_datas:
            for dtype in ("tensor", "NUMPY"):
                result = EnsureTyped(
                    keys="data", data_type=dtype, dtype=np.float32 if dtype == "NUMPY" else None, device="cpu"
                )({"data": test_data})["data"]
                if dtype == "NUMPY":
                    self.assertEqual(result.dtype, np.float32)
                self.assertIsInstance(result, torch.Tensor if dtype == "tensor" else np.ndarray)
                assert_allclose(result, test_data, type_test=False)
                self.assertTupleEqual(result.shape, (2, 2))

    def test_single_input(self):
        test_datas = [5, 5.0, False, np.asarray(5), torch.tensor(5)]
        if torch.cuda.is_available():
            test_datas.append(test_datas[-1].cuda())
        for test_data in test_datas:
            for dtype in ("tensor", "numpy"):
                result = EnsureTyped(keys="data", data_type=dtype)({"data": test_data})["data"]
                self.assertIsInstance(result, torch.Tensor if dtype == "tensor" else np.ndarray)
                if isinstance(test_data, bool):
                    self.assertFalse(result)
                else:
                    assert_allclose(result, test_data, type_test=False)
                self.assertEqual(result.ndim, 0)

    def test_string(self):
        for dtype in ("tensor", "numpy"):
            # string input
            result = EnsureTyped(keys="data", data_type=dtype)({"data": "test_string"})["data"]
            self.assertIsInstance(result, str)
            self.assertEqual(result, "test_string")
            # numpy array of string
            result = EnsureTyped(keys="data", data_type=dtype)({"data": np.array(["test_string"])})["data"]
            self.assertIsInstance(result, np.ndarray)
            self.assertEqual(result[0], "test_string")

    def test_list_tuple(self):
        for dtype in ("tensor", "numpy"):
            result = EnsureTyped(keys="data", data_type=dtype, wrap_sequence=False, track_meta=True)(
                {"data": [[1, 2], [3, 4]]}
            )["data"]
            self.assertIsInstance(result, list)
            self.assertIsInstance(result[0][1], MetaTensor if dtype == "tensor" else np.ndarray)
            assert_allclose(result[1][0], torch.as_tensor(3), type_test=False)
            # tuple of numpy arrays
            result = EnsureTyped(keys="data", data_type=dtype, wrap_sequence=False)(
                {"data": (np.array([1, 2]), np.array([3, 4]))}
            )["data"]
            self.assertIsInstance(result, tuple)
            self.assertIsInstance(result[0], torch.Tensor if dtype == "tensor" else np.ndarray)
            assert_allclose(result[1], torch.as_tensor([3, 4]), type_test=False)

    def test_dict(self):
        # simulate complicated input data
        test_data = {
            "img": np.array([1.0, 2.0], dtype=np.float32),
            "meta": {"dims": 3, "size": np.array([1, 2, 3]), "path": "temp/test"},
            "extra": None,
        }
        for dtype in ("tensor", "numpy"):
            trans = EnsureTyped(keys=["data", "label"], data_type=dtype, dtype=[np.float32, np.int8], device="cpu")(
                {"data": test_data, "label": test_data}
            )
            for key in ("data", "label"):
                result = trans[key]
                self.assertIsInstance(result, dict)
                self.assertIsInstance(result["img"], torch.Tensor if dtype == "tensor" else np.ndarray)
                self.assertIsInstance(result["meta"]["size"], torch.Tensor if dtype == "tensor" else np.ndarray)
                self.assertEqual(result["meta"]["path"], "temp/test")
                self.assertEqual(result["extra"], None)
                assert_allclose(result["img"], torch.as_tensor([1.0, 2.0]), type_test=False)
                assert_allclose(result["meta"]["size"], torch.as_tensor([1, 2, 3]), type_test=False)
            if dtype == "numpy":
                self.assertEqual(trans["data"]["img"].dtype, np.float32)
                self.assertEqual(trans["label"]["img"].dtype, np.int8)
            else:
                self.assertEqual(trans["data"]["img"].dtype, torch.float32)
                self.assertEqual(trans["label"]["img"].dtype, torch.int8)


if __name__ == "__main__":
    unittest.main()



# Copyright (c) MONAI Consortium
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import json
import os
import shutil
import tempfile
import unittest

import numpy as np
from parameterized import parameterized

from monai.data import DataLoader, Dataset
from monai.transforms import Compose, LoadImage, SaveImage, WriteFileMapping
from monai.utils import optional_import

nib, has_nib = optional_import("nibabel")


def create_input_file(temp_dir, name):
    test_image = np.random.rand(128, 128, 128)
    output_ext = ".nii.gz"
    input_file = os.path.join(temp_dir, name + output_ext)
    nib.save(nib.Nifti1Image(test_image, np.eye(4)), input_file)
    return input_file


def create_transform(temp_dir, mapping_file_path, savepath_in_metadict=True):
    return Compose(
        [
            LoadImage(image_only=True),
            SaveImage(output_dir=temp_dir, output_ext=".nii.gz", savepath_in_metadict=savepath_in_metadict),
            WriteFileMapping(mapping_file_path=mapping_file_path),
        ]
    )


@unittest.skipUnless(has_nib, "nibabel required")
class TestWriteFileMapping(unittest.TestCase):
    def setUp(self):
        self.temp_dir = tempfile.mkdtemp()

    def tearDown(self):
        shutil.rmtree(self.temp_dir)

    @parameterized.expand([(True,), (False,)])
    def test_mapping_file(self, savepath_in_metadict):
        mapping_file_path = os.path.join(self.temp_dir, "mapping.json")
        name = "test_image"
        input_file = create_input_file(self.temp_dir, name)
        output_file = os.path.join(self.temp_dir, name, name + "_trans.nii.gz")

        transform = create_transform(self.temp_dir, mapping_file_path, savepath_in_metadict)

        if savepath_in_metadict:
            transform(input_file)
            self.assertTrue(os.path.exists(mapping_file_path))
            with open(mapping_file_path) as f:
                mapping_data = json.load(f)
            self.assertEqual(len(mapping_data), 1)
            self.assertEqual(mapping_data[0]["input"], input_file)
            self.assertEqual(mapping_data[0]["output"], output_file)
        else:
            with self.assertRaises(RuntimeError) as cm:
                transform(input_file)
            cause_exception = cm.exception.__cause__
            self.assertIsInstance(cause_exception, KeyError)
            self.assertIn(
                "Missing 'saved_to' key in metadata. Check SaveImage argument 'savepath_in_metadict' is True.",
                str(cause_exception),
            )

    def test_multiprocess_mapping_file(self):
        num_images = 50

        single_mapping_file = os.path.join(self.temp_dir, "single_mapping.json")
        multi_mapping_file = os.path.join(self.temp_dir, "multi_mapping.json")

        data = [create_input_file(self.temp_dir, f"test_image_{i}") for i in range(num_images)]

        # single process
        single_transform = create_transform(self.temp_dir, single_mapping_file)
        single_dataset = Dataset(data=data, transform=single_transform)
        single_loader = DataLoader(single_dataset, batch_size=1, num_workers=0, shuffle=True)
        for _ in single_loader:
            pass

        # multiple processes
        multi_transform = create_transform(self.temp_dir, multi_mapping_file)
        multi_dataset = Dataset(data=data, transform=multi_transform)
        multi_loader = DataLoader(multi_dataset, batch_size=4, num_workers=3, shuffle=True)
        for _ in multi_loader:
            pass

        with open(single_mapping_file) as f:
            single_mapping_data = json.load(f)
        with open(multi_mapping_file) as f:
            multi_mapping_data = json.load(f)

        single_set = {(entry["input"], entry["output"]) for entry in single_mapping_data}
        multi_set = {(entry["input"], entry["output"]) for entry in multi_mapping_data}

        self.assertEqual(single_set, multi_set)


if __name__ == "__main__":
    unittest.main()
